/*
 * Copyright (c) 2022-2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Sep  7 2022
 */

#include "accamera.h"

#include <err.h>
#include <strings.h>

#include "camera_c_types.h"
#include "codels.hh"

#ifdef HAVE_OPENCV
# include <opencv2/core.hpp>
# include <opencv2/imgcodecs.hpp>
# include <opencv2/imgproc.hpp>
# include <opencv2/videoio.hpp>
#endif


static genom_event	camfcn_record(const char input[16], uint16_t width,
                                uint16_t height, uint16_t fps,
                                const char fourcc[4], const char output[128],
                                const genom_context self);


/* --- Activity publish ------------------------------------------------- */

/** Validation codel camfcn_bank_validate of activity publish.
 *
 * Returns genom_ok.
 * Throws camera_einval.
 */
genom_event
camfcn_bank_validate(const char bank[16], uint8_t index,
                     const genom_context self)
{
  if (!*bank)
    return camera_einval(camera_einval_detail{"empty bank name"}, self);
  if (index >= or_image_max_images)
    return camera_einval(camera_einval_detail{"empty bank name"}, self);

  return genom_ok;
}


/* --- Activity unpublish ----------------------------------------------- */

/** Validation codel camfcn_bank_validate of activity unpublish.
 *
 * Returns genom_ok.
 * Throws camera_einval.
 */
/* already defined in service publish validation */



/* --- Function jpeg ---------------------------------------------------- */

/** Codel camfcn_jpeg of function jpeg.
 *
 * Returns genom_ok.
 * Throws camera_einval.
 */
genom_event
camfcn_jpeg(const camera_worker_lock *mutex, const char input[16],
            uint16_t quality, const char output[16],
            const genom_context self)
{
#ifdef HAVE_OPENCV
  struct camfcn_jpeg : camera_job {
    camfcn_jpeg(int quality) : camera_job(), quality(quality) {}

    void destroy() { delete this; }
    bool runnable() const { return all_input_recent(); }

    state operator()() {
      or_image_frame *src = in(0) ? in(0)->image : nullptr;
      or_image_frame *dst = out(0) ? out(0)->image : nullptr;

      if (!src || !dst) return idle;
      if (!src->pixmap.data._buffer) return idle;

      bool e;
      int type;
      switch(src->pixmap.pixfmt) {
        case or_sensor_pixmap_MONO8: type = CV_8UC1; break;
        case or_sensor_pixmap_MONO16: type = CV_16UC1; break;
        case or_sensor_pixmap_RGB8:
        case or_sensor_pixmap_BGR8: type = CV_8UC3; break;
        case or_sensor_pixmap_RGBA8:
        case or_sensor_pixmap_BGRA8: type = CV_8UC4; break;

        default:
          if (errs.empty())
            errs = camera_pixfmt_str(src->pixmap.pixfmt) +
                   " input format is invalid";
          return idle;
      }
      cv::Mat s(
        src->pixmap.height, src->pixmap.width, type,
        src->pixmap.data._buffer, src->pixmap.stride);
      cv::Mat *current = &s;

      switch(src->pixmap.pixfmt) {
        case or_sensor_pixmap_RGB8:
          cv::cvtColor(*current, cvt, cv::COLOR_RGB2BGR);
          current = &cvt;
          break;

        case or_sensor_pixmap_RGBA8:
          cv::cvtColor(*current, cvt, cv::COLOR_RGBA2BGRA);
          current = &cvt;
          break;
      }
      e = cv::imencode(
        ".jpeg", *current, jpeg, { cv::IMWRITE_JPEG_QUALITY, quality });
      if (!e) return idle;

      dst->ts = src->ts;
      if (dst->info.seq != src->info.seq)
        dst->info = src->info;

      dst->pixmap.seq = src->pixmap.seq;
      dst->pixmap.width = src->pixmap.width;
      dst->pixmap.height = src->pixmap.height;
      dst->pixmap.stride = src->pixmap.stride;
      dst->pixmap.pixfmt = or_sensor_pixmap_JPEG;
      dst->pixmap.data._maximum = jpeg.size();
      dst->pixmap.data._length = dst->pixmap.data._maximum;
      dst->pixmap.data._buffer = jpeg.data();
      dst->pixmap.data._release = NULL;
      errs.clear();
      return update;
    }

   private:
    int quality;
    cv::Mat cvt;
    std::vector<unsigned char> jpeg;
  };

  camfcn_jpeg *filter = new camfcn_jpeg(quality);
  filter->sink(output);
  filter->src(input);
  return genom_ok;

#else
  (void)input, (void)quality, (void)output;
  return camera_einval(camera_einval_detail{"not supported"}, self);
#endif
}


/* --- Function record_mpeg --------------------------------------------- */

/** Codel camfcn_record_mpeg of function record_mpeg.
 *
 * Returns genom_ok.
 * Throws camera_einval.
 */
genom_event
camfcn_record_mpeg(const camera_worker_lock *mutex,
                   const char input[16], uint16_t width,
                   uint16_t height, uint16_t fps,
                   const char output[128], const genom_context self)
{
  /* opencv insists on making an AVI container */
  const char *dot = strrchr(output, '.');
  if (dot && strcasecmp(dot + 1, "avi"))
    return camera_einval(
      camera_einval_detail{".avi output extension required"}, self);

  return camfcn_record(input, width, height, fps, "MPEG", output, self);
}


/* --- camfcn_record ------------------------------------------------------- */

/* generic recording function */
static genom_event
camfcn_record(const char input[16], uint16_t width, uint16_t height,
              uint16_t fps, const char fourcc[4], const char output[128],
              const genom_context self)
{
#ifdef HAVE_OPENCV
  struct camfcn_record : camera_job {
    camfcn_record(const char *output, uint16_t width, uint16_t height,
                  const char fourcc[4], uint16_t fps) :
      camera_job(), size(width, height),
      writer(
        output,
        cv::VideoWriter::fourcc(fourcc[0], fourcc[1], fourcc[2], fourcc[3]),
        fps, size) {
      if (!writer.isOpened())
        CV_Error(cv::Error::StsBadArg, "invalid arguments");
    }

    void destroy() { delete this; }
    bool runnable() const { return all_input_recent(); }

    state operator()() {
      or_image_frame *src = in(0) ? in(0)->image : nullptr;
      or_image_frame *dst = out(0) ? out(0)->image : nullptr;
      int type;

      if (!src || !src->pixmap.data._buffer) return idle;

      /* check input format */
      switch(src->pixmap.pixfmt) {
        case or_sensor_pixmap_MONO8: type = CV_8UC1; break;
        case or_sensor_pixmap_MONO16: type = CV_16UC1; break;
        case or_sensor_pixmap_RGB8:
        case or_sensor_pixmap_BGR8: type = CV_8UC3; break;
        case or_sensor_pixmap_RGBA8:
        case or_sensor_pixmap_BGRA8: type = CV_8UC4; break;

        default:
          if (errs.empty())
            errs = camera_pixfmt_str(src->pixmap.pixfmt) +
                   " input format is invalid";
          return idle;
      }
      cv::Mat s(
        src->pixmap.height, src->pixmap.width, type,
        src->pixmap.data._buffer, src->pixmap.stride);
      cv::Mat *current = &s;

      /* resize and convert colorspace */
      if (s.size() != size) {
        cv::resize(*current, scaled, size, 0, 0, cv::INTER_NEAREST);
        current = &scaled;
      }
      switch(src->pixmap.pixfmt) {
        case or_sensor_pixmap_RGB8:
          cv::cvtColor(*current, color, cv::COLOR_RGB2BGR);
          current = &color;
          break;

        case or_sensor_pixmap_RGBA8:
          cv::cvtColor(*current, color, cv::COLOR_RGBA2BGR);
          current = &color;
          break;
      }

      /* encode to output file */
      writer << *current;

      /* output slot */
      if (!dst) return update;

      dst->ts = src->ts;
      if (dst->info.seq != src->info.seq)
        dst->info = src->info;

      dst->pixmap.seq = src->pixmap.seq;
      dst->pixmap.width = current->cols;
      dst->pixmap.height = current->rows;
      dst->pixmap.stride = current->step;
      dst->pixmap.pixfmt = or_sensor_pixmap_BGR8;
      dst->pixmap.data._maximum = current->step * current->rows;
      dst->pixmap.data._length = dst->pixmap.data._maximum;
      dst->pixmap.data._buffer = current->data;
      dst->pixmap.data._release = NULL;
      errs.clear();
      return update;
    }

   private:
    cv::Size size;
    cv::Mat scaled, color;
    cv::VideoWriter writer;
  };

  try {
    camfcn_record *filter =
      new camfcn_record(output, width, height, fourcc, fps);

    /* create a sink slot named after output basename */
    const char *slash = strrchr(output, '/');
    slash = slash ? slash + 1 : output;
    filter->sink(slash);

    /* connect to the source */
    filter->src(input);

  } catch(std::exception &e) {
    camera_einval_detail d;
    snprintf(d.what, sizeof(d.what), "%s", e.what());
    return camera_einval(d, self);
  }

  return genom_ok;

#else
  (void)input, (void)quality, (void)output;
  return camera_einval(camera_einval_detail{"not supported"}, self);
#endif
}


/* --- Function cancel -------------------------------------------------- */

/** Codel camfcn_cancel of function cancel.
 *
 * Returns genom_ok.
 * Throws camera_einval.
 */
genom_event
camfcn_cancel(const camera_worker_lock *mutex, const char slot[16],
              const genom_context self)
{
  camera_slot *s = camera_slot::get(slot);
  if (!s)
    return camera_einval(camera_einval_detail{"not such slot"}, self);

  s->cancel();
  return genom_ok;
}


/* --- Function status -------------------------------------------------- */

/** Codel camfcn_status of function status.
 *
 * Returns genom_ok.
 * Throws camera_einval.
 */
genom_event
camfcn_status(const camera_worker_lock *mutex, const char slot[16],
              const genom_context self)
{
  camera_einval_detail d;
  genom_event e;

  camera_slot *s = camera_slot::get(slot);
  if (!s)
    return camera_einval(camera_einval_detail{"not such slot"}, self);

  e = s->status(d);
  return e == genom_ok ? genom_ok : self->raise(e, &d, sizeof(d), self);
}
