/*
 * Copyright (c) 2022-2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Sep  7 2022
 */

#include "accamera.h"

#include <err.h>
#include <unistd.h>

#include <typeinfo>

#include "camera_c_types.h"
#include "codels.hh"


/* --- Task worker ------------------------------------------------------ */


/** Codel camwrk_start of task worker.
 *
 * Triggered by camera_start.
 * Yields to camera_poll.
 */
genom_event
camwrk_start(camera_worker *worker, const genom_context self)
{
  /* only one worker has to initialize the context */
  if (worker->pipeline) return camera_poll;

  worker->pipeline = new camera_pipeline();
  return camera_poll;
}


/** Codel camwrk_poll of task worker.
 *
 * Triggered by camera_poll.
 * Yields to camera_poll, camera_process, camera_publish,
 *           camera_ether.
 */
genom_event
camwrk_poll(const camera_pipeline *pipeline, camera_pub **pub,
            camera_job **job, const genom_context self)
{
  if (!pipeline) return camera_ether;

  if (pipeline->poll(pub, job, 500/*ms*/)) {
    if (*pub) return camera_publish;
    if (*job) return camera_process;
  }

  return camera_poll;
}


/** Codel camwrk_process of task worker.
 *
 * Triggered by camera_process.
 * Yields to camera_poll.
 */
genom_event
camwrk_process(const camera_pipeline *pipeline, uint16_t *task,
               camera_job **job, const genom_context self)
{
  if (!pipeline) return camera_poll;

  auto s = pipeline->run(*job);
  pipeline->fini(*job, s);
  return camera_poll;
}


/** Codel camwrk_publish of task worker.
 *
 * Triggered by camera_publish.
 * Yields to camera_poll.
 */
genom_event
camwrk_publish(const camera_pipeline *pipeline, uint16_t *task,
               camera_pub **pub, const camera_image *image,
               const genom_context self)
{
  or_image_bank *port = image->data((*pub)->bank.c_str(), self);
  if (port == NULL) return camera_poll;
  if (!pipeline) return camera_poll;

  if (pipeline->run(*pub, port))
    image->write((*pub)->bank.c_str(), self);

  pipeline->fini(*pub);

  return camera_poll;
}


/** Codel camwrk_stop of task worker.
 *
 * Triggered by camera_stop.
 * Yields to camera_ether.
 */
genom_event
camwrk_stop(camera_worker *worker, const genom_context self)
{
  /* only one worker has to finalize the context */
  if (!worker->pipeline) return camera_ether;

  delete worker->pipeline;
  worker->pipeline = nullptr;
  return camera_ether;
}


/* --- Activity publish ------------------------------------------------- */

/** Codel camact_publish of activity publish.
 *
 * Triggered by camera_start.
 * Yields to camera_ether.
 * Throws camera_einval.
 */
genom_event
camact_publish(const camera_pipeline *pipeline,
               const camera_worker_lock *mutex, const char slot[16],
               const char bank[16], uint8_t index,
               const camera_image *image, const genom_context self)
{
  or_image_bank *port = image->data(bank, self);
  bool clean = false;

  /* create port if needed */
  if (port == NULL) {
    clean = true;
    if (image->open(bank, self) || (port = image->data(bank, self)) == NULL)
      return camera_einval(camera_einval_detail{"invalid bank name"}, self);
    *port = {};
  }

  /* reserve images */
  if (port->images._length < index + 1U) {
    while(port->images._length <= index)
      port->images._buffer[port->images._length++] = {};
    if (image->write(bank, self)) {
      if (clean) image->close(bank, self);
      return camera_eio(camera_eio_detail{"cannot write bank"}, self);
    }
  }

  /* update publisher */
  camera_pub *p = camera_pub::get(bank);
  if (!p) {
    if (clean) image->close(bank, self);
    return camera_einval(camera_einval_detail{"invalid bank name"}, self);
  }

  /* add source */
  p->src(index, slot);
  return camera_ether;
}


/* --- Activity unpublish ----------------------------------------------- */

/** Codel camact_unpublish of activity unpublish.
 *
 * Triggered by camera_start.
 * Yields to camera_ether.
 * Throws camera_einval.
 */
genom_event
camact_unpublish(const camera_pipeline *pipeline,
                 const camera_worker_lock *mutex, const char bank[16],
                 uint8_t index, const camera_image *image,
                 const genom_context self)
{
  or_image_bank *port = image->data(bank, self);

  if (!port)
    return camera_einval(camera_einval_detail{"no such bank"}, self);
  if (port->images._length < index + 1U)
    return camera_einval(camera_einval_detail{"no such index"}, self);

  /* update publisher */
  camera_pub *p = camera_pub::get(bank);
  if (p) p->reset(index);

  /* update index */
  port->images._buffer[index] = {};
  image->write(bank, self);
  return camera_ether;
}
