/*
 * Copyright (c) 2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Dec 14 2023
 */
#ifndef HH_CAMERA_CODELS_GENOM3
#define HH_CAMERA_CODELS_GENOM3

#include <sys/time.h>

#include <condition_variable>
#include <functional>
#include <mutex>
#include <string>
#include <unordered_map>

#include "camera_c_types.h"
#include "camera.h"


/* --- publisher ----------------------------------------------------------- */

struct camera_pub {
  camera_pub(const std::string &bank) : bank(bank) {}
  camera_pub(camera_pub &) = delete;
  void operator=(const camera_pub &) = delete;

  camera_pub *src(uint8_t index, const std::string &slot);
  void reset(uint8_t index);

  bool runnable();
  const camera_slot_image *in(size_t index) const;

  static camera_pub *get(const std::string &bank);

  const std::string bank;

 private:
  typedef std::unordered_map<std::string, camera_pub> bankmap;
  friend camera_pipeline;

  bool lock();
  void unlock();
  bool operator()(const camera_pipeline *pipeline, or_image_bank *port);

  struct timespec ts{0, 0};
  struct { camera_job::slots in{or_image_max_images}; } current, locked;
  camera_pub **ref{nullptr};

  static bankmap banks;
};


/* --- slots --------------------------------------------------------------- */

struct camera_slot {
  typedef std::vector<camera_job *> jobs;
  typedef std::vector<camera_pub *> pubs;

  camera_slot(const std::string &name)
    : name(name), image{new or_image_frame()} {}

  genom_event status(camera_einval_detail &detail) const;
  const struct timespec &ts() const { return image.ts; }
  struct timespec &ts() { return image.ts; }
  const camera_slot_image *simage() const { return &image; }

  bool rdlock();
  bool wrlock();
  void unlock();

  const pubs &pubrds() const { return pubrd; };
  const jobs &jobrds() const { return jobrd; };
  const jobs &jobwrs() const { return jobwr; };

  void cancel();
  void remove(const camera_pub *pub);
  void remove(const camera_job *job);

  static camera_slot *src(const std::string &name, camera_pub *pub);
  static camera_slot *src(const std::string &name, camera_job *job);
  static camera_slot *sink(const std::string &name, camera_job *job);
  static camera_slot *get(const std::string &name);

  const std::string name;

 private:
  camera_slot_image image;

  pubs pubrd{};
  jobs jobrd{}, jobwr{};
  int lock{0};

  static std::unordered_map<std::string, camera_slot> byname;
};


/* --- pipeline ------------------------------------------------------------ */

struct camera_pipeline {
  camera_pipeline();
  ~camera_pipeline();

  int poll(camera_pub **pub, camera_job **job, int tout) const;

  bool schedule(camera_job *job) const;
  bool run(camera_pub *pub, or_image_bank *port) const;
  camera_job::state run(camera_job *job) const;
  void fini(camera_pub *pub) const;
  void fini(camera_job *job, camera_job::state s) const;

  static std::mutex mutex;

 private:
  template<typename T> struct queue;

  void reschedule(camera_job::slots *in, camera_job::slots *out) const;

  mutable std::condition_variable cv;
  mutable bool publishing{false};

  queue<camera_pub> *pubq;
  queue<camera_job> *jobq;
};


/* helpers */

static inline std::string
camera_pixfmt_str(or_sensor_pixmap_encoding fmt)
{
  switch(fmt) {
    case or_sensor_pixmap_MONO8: return "MONO8";
    case or_sensor_pixmap_MONO16: return "MONO16";
    case or_sensor_pixmap_Z32: return "Z32";
    case or_sensor_pixmap_XYZ32: return "XYZ32";
    case or_sensor_pixmap_JPEG: return "JPEG";
  }
  return "INVALID";
}

static inline genom_event
camera_einval(const camera_einval_detail &d, genom_context self) {
  return camera_einval(&d, self);
}

static inline genom_event
camera_eio(const camera_eio_detail &d, genom_context self) {
  return camera_eio(&d, self);
}


#endif /* HH_CAMERA_CODELS_GENOM3 */
