/*
 * Copyright (c) 2023-2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Nov 22 2023
 */
#include "accamera.h"

#include <err.h>

#include <stdexcept>

#include "codels.hh"


/* --- camera_job::~camera_job --------------------------------------------- */

camera_job::~camera_job()
{
  reset();

  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);
  unlock(true);
  if (ref) { *ref = nullptr; ref = nullptr; }
}


/* --- camera_job::status -------------------------------------------------- */

genom_event
camera_job::status(camera_einval_detail &detail) const
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);

  if (errs.empty()) return genom_ok;

  snprintf(detail.what, sizeof(detail.what), "%s", errs.c_str());
  return camera_einval_id;
}


/* --- camera_job::src ----------------------------------------------------- */

camera_slot *
camera_job::src(const std::string &name)
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);

  camera_slot *s = camera_slot::src(name, this);
  current.in.push_back(s);
  return s;
}


/* --- camera_job::sink ---------------------------------------------------- */

camera_slot *
camera_job::sink(const std::string &name)
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);

  camera_slot *s = camera_slot::sink(name, this);
  current.out.push_back(s);
  return s;
}


/* --- camera_job::reset --------------------------------------------------- */

void
camera_job::reset()
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);

  for(auto s : current.in) if (s) s->remove(this);
  for(auto s : current.out) if (s) s->remove(this);

  current.in.clear();
  current.out.clear();
}


/* --- camera_job::schedule ------------------------------------------------ */

bool
camera_job::schedule(const camera_pipeline *pipeline)
{
  return pipeline ? pipeline->schedule(this) : false;
}


/* --- camera_job::all_input_recent ---------------------------------------- */

bool
camera_job::all_input_recent() const
{
  for(auto s : current.in) {
    if (!s) continue;

    auto &sts = s->ts();
    if (sts.tv_sec < ts.tv_sec) return false;
    if (sts.tv_sec == ts.tv_sec && sts.tv_nsec < ts.tv_nsec) return false;
  }

  return current.in.size();
}


/* --- camera_job slots ---------------------------------------------------- */

size_t
camera_job::nin() const
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);
  return current.in.size();
}

const camera_slot_image *
camera_job::in(size_t index) const
{
  if (index >= current.in.size()) return nullptr;
  if (index >= locked.in.size()) return nullptr;

  return locked.in[index] == current.in[index] ?
    locked.in[index]->simage() : nullptr;
}

size_t
camera_job::nout() const
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);
  return current.out.size();
}

const camera_slot_image *
camera_job::out(size_t index) const
{
  if (index >= current.out.size()) return nullptr;
  if (index >= locked.out.size()) return nullptr;

  return locked.out[index] == current.out[index] ?
    locked.out[index]->simage() : nullptr;
}


/* --- camera_job lock ----------------------------------------------------- */

bool
camera_job::lock()
{
  for(auto s = current.in.begin(); s != current.in.end(); ++s) {
    if (!*s) continue;
    if (!(*s)->rdlock()) {
      for(auto i = current.in.begin(); i != s; i++)
        if (*i) (*i)->unlock();
      return false;
    }
  }

  for(auto s = current.out.begin(); s != current.out.end(); ++s) {
    if (!*s) continue;
    if (!(*s)->wrlock()) {
      for(auto i = current.out.begin(); i != s; i++)
        if (*i) (*i)->unlock();
      for(auto s : current.in) if (s) s->unlock();
      return false;
    }
  }

  locked = current;
  return true;
}

void
camera_job::unlock(bool update)
{
  if (update) {
    clock_gettime(CLOCK_MONOTONIC, &ts);
    for(auto s : locked.out) if (s) s->ts() = ts;
  }

  for(auto s : locked.out) if (s) s->unlock();
  for(auto s : locked.in) if (s) s->unlock();
  locked.out.clear();
  locked.in.clear();
}


/* === camera_publisher ==================================================== */

camera_pub::bankmap camera_pub::banks;	/* per-bank jobs */

/* --- camera_pub::get ----------------------------------------------------- */

camera_pub *
camera_pub::get(const std::string &bank)
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);

  try { return &banks.at(bank); }
  catch (std::out_of_range &) {
    auto i = banks.emplace(bank, bank);
    return &i.first->second;
  }
}


/* --- camera_pub::src ----------------------------------------------------- */

/* register a new slot for publication at the given index */
camera_pub *
camera_pub::src(uint8_t index, const std::string &slot)
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);

  if (current.in[index]) current.in[index]->remove(this);
  current.in[index] = camera_slot::src(slot, this);
  return this;
}


/* --- camera_pub::reset --------------------------------------------------- */

/* remove the given index */
void
camera_pub::reset(uint8_t index)
{
  std::lock_guard<std::mutex> lk(camera_pipeline::mutex);

  if (!current.in[index]) return;
  current.in[index]->remove(this);
  current.in[index] = nullptr;
}


/* --- camera_pub::runnable ------------------------------------------------ */

/* true if all slots of a bank have been updated since last publication */
bool
camera_pub::runnable()
{
  bool up = false;

  for(auto s : current.in) {
    if (!s) continue;
    up = true;

    const auto &sts = s->ts();
    if (sts.tv_sec < ts.tv_sec) return false;
    if (sts.tv_sec == ts.tv_sec && sts.tv_nsec < ts.tv_nsec) return false;
  }

  return up;
}


/* --- camera_pub slots ---------------------------------------------------- */

const camera_slot_image *
camera_pub::in(size_t index) const
{
  if (index >= locked.in.size()) return nullptr;
  if (!locked.in[index]) return nullptr;

  return locked.in[index] == current.in[index] ?
    locked.in[index]->simage() : nullptr;
}


/* --- camera_pub::lock / unlock ------------------------------------------- */

bool
camera_pub::lock()
{
  for(auto s = current.in.begin(); s != current.in.end(); ++s) {
    if (!*s) continue;
    if (!(*s)->rdlock()) {
      for(auto i = current.in.begin(); i != s; i++)
        if (*i) (*i)->unlock();
      return false;
    }
  }

  locked = current;
  return true;
}

void
camera_pub::unlock()
{
  for(auto &s : locked.in) if (s) { s->unlock(); s = nullptr; }
}


/* --- camera_pub::operator() ---------------------------------------------- */

bool
camera_pub::operator()(const camera_pipeline *pipeline, or_image_bank *port)
{
  or_image_frame *dst = port->images._buffer;
  const camera_slot_image *src;
  bool fresh = false;

  for(auto i = 0; i < or_image_max_images; ++i, ++dst) {
    src = in(i);
    if (!src) continue;

    if (dst->ts.nsec != src->image->ts.nsec ||
        dst->ts.sec != src->image->ts.sec) {
      fresh = true;
      dst->ts = src->image->ts;
    }

    if (dst->info.seq != src->image->info.seq) {
      fresh = true;
      dst->info = src->image->info;
    }

    if (dst->pixmap.seq != src->image->pixmap.seq) {
      fresh = true;
      dst->pixmap = src->image->pixmap;
      dst->pixmap.data._release = NULL; /* data belongs to camera_slot */
    }

    if (port->ts.sec < dst->ts.sec
        || (port->ts.sec == dst->ts.sec && port->ts.nsec < dst->ts.nsec)) {
      fresh = true;
      port->ts.sec = dst->ts.sec;
      port->ts.nsec = dst->ts.nsec;
    }
  }

  clock_gettime(CLOCK_MONOTONIC, &ts);
  return fresh;
}
