/*
 * Copyright (c) 2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Nov 22 2023
 */
#include "accamera.h"

#include <err.h>

#include <chrono>
#include <functional>
#include <memory>
#include <mutex>
#include <string>
#include <unordered_map>
#include <vector>

#include "codels.hh"


template<typename T>
struct camera_pipeline::queue {
  void enqueue(T *data, std::condition_variable &cv);
  bool dequeue(T **data);
  void remove(T *data);

 private:
  struct node { T *data; node *next{nullptr}; } *head{nullptr};
  static node *pool;
};

template<typename T> typename camera_pipeline::queue<T>::node *
camera_pipeline::queue<T>::pool = nullptr;


/* --- camera_pipeline ----------------------------------------------------- */

std::mutex camera_pipeline::mutex;

camera_pipeline::camera_pipeline() :
  pubq(new queue<camera_pub>), jobq(new queue<camera_job>)
{
}

camera_pipeline::~camera_pipeline()
{
  std::unique_lock<std::mutex> lk(mutex);

  delete pubq;
  delete jobq;
}


/* --- camera_pipeline::poll ----------------------------------------------- */

int
camera_pipeline::poll(camera_pub **pub, camera_job **job, int tout) const
{
  auto dl =
    std::chrono::system_clock::now() + std::chrono::milliseconds(tout);

  std::unique_lock<std::mutex> lk(mutex);
  *pub = nullptr;
  *job = nullptr;

  while(1) {
    if (!publishing && pubq->dequeue(pub)) {
      if (!*pub) continue;
      publishing = true;
      return 1;
    }

    if (jobq->dequeue(job)) {
      if (!*job) continue;
      (*job)->ref = job; /* remember where the reference is handed over */
      return 1;
    }

    if (tout == 0) return 0;
    if (cv.wait_until(lk, dl) == std::cv_status::timeout)
      return 0;
  }
}


/* --- camera_pipeline::schedule ------------------------------------------- */

bool
camera_pipeline::schedule(camera_job *job) const
{
  std::lock_guard<std::mutex> lk(mutex);

  if (!job->runnable() || !job->lock()) return false;
  jobq->enqueue(job, cv);
  return true;
}


/* --- camera_pipeline::run ------------------------------------------------ */

bool
camera_pipeline::run(camera_pub *pub, or_image_bank *bank) const
{
  return pub ? (*pub)(this, bank) : false;
}


camera_job::state
camera_pipeline::run(camera_job *job) const
{
  return job ? (*job)() : camera_job::cancel;
}


/* --- camera_pipeline::fini ----------------------------------------------- */

void
camera_pipeline::fini(camera_pub *pub) const
{
  std::lock_guard<std::mutex> lk(mutex);
  publishing = false;
  if (!pub) return;

  pub->unlock();
  reschedule(&pub->current.in, nullptr);
}

void
camera_pipeline::fini(camera_job *job, camera_job::state s) const
{
  if (!job) return;

  {
    std::lock_guard<std::mutex> lk(mutex);

    job->unlock(s != camera_job::idle);
    if (job->ref) { *job->ref = nullptr; job->ref = nullptr; }
    reschedule(&job->current.in, &job->current.out);
  }

  if (s == camera_job::cancel) job->destroy();
}


/* --- camera_pipeline::reschedule ----------------------------------------- */

void
camera_pipeline::reschedule(camera_job::slots *in,
                            camera_job::slots *out) const
{
  if (out) {
    for(auto s : *out) {
      if (!s) continue;
      for(auto p : s->pubrds())
        if (p->runnable() && p->lock()) pubq->enqueue(p, cv);
    }
    for(auto s : *out) {
      if (!s) continue;
      for(auto j : s->jobrds())
        if (j->runnable() && j->lock()) jobq->enqueue(j, cv);
    }
  }
  if (in) {
    for(auto s : *in) {
      if (!s) continue;
      for(auto j : s->jobwrs())
        if (j->runnable() && j->lock()) jobq->enqueue(j, cv);
    }
  }
}


/* --- camera_pipeline::queue ---------------------------------------------- */

template<typename T> void
camera_pipeline::queue<T>::enqueue(T *data, std::condition_variable &cv)
{
  node *n, **tail;

  for(tail = &head; *tail; tail = &(*tail)->next)
    if ((*tail)->data == data) return;

  n = pool ? pool : new node;
  pool = n->next;
  n->data = data;
  n->next = nullptr;
  *tail = n;

  data->ref = &n->data;
  cv.notify_one();
}

template<typename T> bool
camera_pipeline::queue<T>::dequeue(T **data)
{
  node *n = head;

  if (!n) return false;

  *data = n->data;
  head = head->next;
  n->next = pool;
  pool = n;

  (*data)->ref = nullptr;
  return true;
}

template<typename T> void
camera_pipeline::queue<T>::remove(T *data)
{
  node *n;

  for(n = head; n; n = n->next)
    if (n->data == data) n->data = nullptr;
}
