/*
 * Copyright (c) 2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Jan 18 2024
 */
#include "accamera.h"

#include <err.h>

#include <algorithm>
#include <vector>
#include <stdexcept>

#include "codels.hh"

std::unordered_map<std::string, camera_slot> camera_slot::byname;


/* --- camera_slot::status ------------------------------------------------- */

genom_event
camera_slot::status(camera_einval_detail &detail) const
{
  genom_event e;

  for (auto j : jobwr) {
    e = j->status(detail);
    if (e != genom_ok) return e;
  }
  return genom_ok;
}


/* --- camera_slot::lock / unlock ------------------------------------------ */

bool
camera_slot::rdlock()
{
  return lock < 0 ? false : ++lock;
}

bool
camera_slot::wrlock()
{
  return lock != 0 ? false : lock = -1;
}

void
camera_slot::unlock()
{
  lock > 0 ? --lock : lock = 0;
}


/* --- camera_slot::cancel ------------------------------------------------- */

void
camera_slot::cancel()
{
  for (auto j : jobwr) j->destroy();
}


/* --- camera_slot::remove ------------------------------------------------- */

void
camera_slot::remove(const camera_pub *pub)
{
  pubrd.erase(std::remove(pubrd.begin(), pubrd.end(), pub), pubrd.end());
}

void
camera_slot::remove(const camera_job *job)
{
  jobrd.erase(std::remove(jobrd.begin(), jobrd.end(), job), jobrd.end());
  jobwr.erase(std::remove(jobwr.begin(), jobwr.end(), job), jobwr.end());
}


/* --- camera_slot::src ---------------------------------------------------- */

camera_slot *
camera_slot::src(const std::string &name, camera_pub *pub)
{
  camera_slot *s = get(name);

  if (std::find(s->pubrd.begin(), s->pubrd.end(), pub) == s->pubrd.end())
    s->pubrd.push_back(pub);

  return s;
}

camera_slot *
camera_slot::src(const std::string &name, camera_job *job)
{
  camera_slot *s = get(name);

  if (std::find(s->jobrd.begin(), s->jobrd.end(), job) == s->jobrd.end())
    s->jobrd.push_back(job);

  return s;
}


/* --- camera_slot::sink --------------------------------------------------- */

camera_slot *
camera_slot::sink(const std::string &name, camera_job *job)
{
  camera_slot *s = get(name);

  if (std::find(s->jobwr.begin(), s->jobwr.end(), job) == s->jobwr.end())
    s->jobwr.push_back(job);

  return s;
}


/* --- camera_slot::get ---------------------------------------------------- */

camera_slot *
camera_slot::get(const std::string &name)
{
  try { return &camera_slot::byname.at(name); }
  catch (std::out_of_range &) {
    auto i = byname.emplace(name, name);
    return &i.first->second;
  }
}
